@@ -10,9 +10,9 @@ from scipy.spatial.distance import squareform
 from scipy.linalg import solve as scipy_solve
 from numpy.linalg import solve as numpy_solve, LinAlgError, inv
 from multiprocessing import Pool
-import scipy.spatial.distance
 
 from .Variogram import Variogram
+from .MetricSpace import MetricSpace, MetricSpacePair
 
 
 class LessPointsError(RuntimeError):
@@ -41,7 +41,10 @@ class OrdinaryKriging:
             precision=100,
             solver='inv',
             n_jobs=1,
-            perf=False
+            perf=False,
+            sparse=False,
+            coordinates=None,
+            values=None
     ):
         """Ordinary Kriging routine
 
@@ -88,14 +91,27 @@ class OrdinaryKriging:
             If True, the different parts of the algorithm will record their
             processing time. This is meant to be used for optimization and
             will be removed in a future version. Do not rely on this argument.
+        sparse : bool
+
+        coordinates: numpy.ndarray, MetricSpace
+        values: numpy.ndarray
 
         """
         # store arguments to the instance
-        if not isinstance(variogram, Variogram):
-            raise TypeError('variogram has to be of type skgstat.Variogram.')
+
+        if isinstance(variogram, Variogram):
+            if coordinates is None:
+                coordinates = variogram.coordinates
+            if values is None:
+                values = variogram.values
+            variogram_descr = variogram.describe()
+            if variogram_descr["model"] == "harmonize":
+                variogram_descr["model"] = variogram._build_harmonized_model()
+            variogram = variogram_descr
+
+        self.sparse = sparse
 
         # general attributes
-        self.V = variogram
         self._minp = min_points
         self._maxp = max_points
         self.min_points = min_points
@@ -105,14 +121,21 @@ class OrdinaryKriging:
         self.n_jobs = n_jobs
         self.perf = perf
 
-        params = self.V.describe()
-        self.range = params['effective_range']
-        self.nugget = params['nugget']
-        self.sill = params['sill']
+        self.range = variogram['effective_range']
+        self.nugget = variogram['nugget']
+        self.sill = variogram['sill']
+        self.dist_metric = variogram["dist_func"]
 
         # coordinates and semivariance function
-        self.coords, self.values = self._get_coordinates_and_values()
-        self.gamma_model = self.V.fitted_model
+        if not isinstance(coordinates, MetricSpace):
+            coordinates, values = self._remove_duplicated_coordinates(coordinates, values)
+            coordinates = MetricSpace(coordinates.copy(), self.dist_metric, self.range if self.sparse else None)
+        else:
+            assert self.dist_metric == coordinates.dist_metric, "Distance metric of variogram differs from distance metric of coordinates"
+            assert coordinates.max_dist is None or coordinates.max_dist == self.range, "Sparse coordinates must have max_dist == variogram.effective_range"
+        self.values = values.copy()
+        self.coords = coordinates
+        self.gamma_model = Variogram.fitted_model_function(**variogram)
         self.z = None
 
         # calculation mode; self.range has to be initialized
@@ -139,33 +162,21 @@ class OrdinaryKriging:
             self.perf_mat = list()
             self.perf_solv = list()
 
-    @property
-    def dist(self):
-        return self.V._dist_func_wrapper
+    def dist(self, x):
+        return Variogram.wrapped_distance_function(self.dist_metric, x)
 
-    @property
-    def dist_metric(self):
-        return self.V._dist_func_name
-               
-    def _get_coordinates_and_values(self):
+    @classmethod
+    def _remove_duplicated_coordinates(cls, coords, values):
         """Extract the coordinates and values
 
-        The coordinates and values array is extracted from the Variogram
-        instance. Additionally, the coordinates array is checked for
-        duplicates and only the first instance of a duplicate is used.
-        Duplicated coordinates would result in duplicated rows in the
-        variogram matrix and make it singular.
-
-        Returns
-        -------
-        coords : numpy.array
-            copy of Variogram.coordines without duplicates
-        values : numpy.array
-            copy of Variogram.values without duplicates
+        The coordinates array is checked for duplicates and only the
+        first instance of a duplicate is used. Duplicated coordinates
+        would result in duplicated rows in the variogram matrix and
+        make it singular.
 
         """
-        c = self.V.coordinates.copy()
-        v = self.V.values.copy()
+        c = coords
+        v = values
 
         _, idx = np.unique(c, axis=0, return_index=True)
 
@@ -258,9 +269,12 @@ class OrdinaryKriging:
         returns an estimation of the observable for the given unobserved
         locations. Each coordinate dimension should be a 1D array.
 
+        .. versionchanged:: 0.6.4
+            sigma array is now initialized with NaN, instead of empty.
+
         Parameters
         ----------
-        x : numpy.array
+        x : numpy.array, MetricSpace
             One 1D array for each coordinate dimension. Typically two or
             three array, x, y, (z) are passed for 2D and 3D Kriging
 
@@ -279,20 +293,25 @@ class OrdinaryKriging:
         if self.perf:
             self.perf_dist, self.perf_mat, self.perf_solv = [], [], []
 
-        self.transform_coordinates = np.column_stack(x)
-        self.transform_dists = scipy.spatial.distance.cdist(self.transform_coordinates, self.coords, metric=self.dist_metric)
-        
+        if len(x) != 1 or not isinstance(x[0], MetricSpace):
+            self.transform_coords = MetricSpace(np.column_stack(x).copy(), self.dist_metric, self.range if self.sparse else None)
+        else:
+            self.transform_coords = x[0]
+        self.transform_coords_pair = MetricSpacePair(self.transform_coords, self.coords)
+
         # DEV: this is dirty, not sure how to do it better at the moment
-        self.sigma = np.empty(len(x[0]))
+        #self.sigma = np.empty(len(x[0]))
+        self.sigma = np.ones(len(x[0])) * np.nan
         self.__sigma_index = 0
+
         # if multi-core, than here
         if self.n_jobs is None or self.n_jobs == 1:
-            z = np.fromiter(map(self._estimator, range(len(self.transform_coordinates))), dtype=float)
+            z = np.fromiter(map(self._estimator, range(len(self.transform_coords))), dtype=float)
         else:
             def f(idxs):
                 return self._estimator(idxs)
             with Pool(self.n_jobs) as p:
-                z = p.starmap(f, range(len(self.transform_coordinates)))
+                z = p.starmap(f, range(len(self.transform_coords)))
 
         # print warnings
         if self.singular_error > 0:
@@ -317,27 +336,41 @@ class OrdinaryKriging:
         LessPointsError are handled and the error counters are increased. In
         both cases numpy.NaN will be used as estimate.
 
+        ..versionchanged:: 0.6.4
+            sigma_index is now always incremented
+
         """
+        # indicate if this esimation raised an error
+        did_error = False
+
+        # reun estimation
         try:
-            z, sigma = self._krige(idx)
+            estimation, sigma = self._krige(idx)
         except SingularMatrixError:
             self.singular_error += 1
-            return np.nan
+            did_error = True
         except LessPointsError:
             self.no_points_error += 1
-            return np.nan
+            did_error = True
         except IllMatrixError:
             self.ill_matrix += 1
-            return np.nan
+            did_error = True
 
         # TODO: This is a side-effect and I need to re-design this part:
-        self.sigma[self.__sigma_index] = sigma
+        if not did_error:
+            self.sigma[self.__sigma_index] = sigma
+            z = estimation
+        else:
+            # on error - Z* is NaN
+            z = np.nan
+
+        # in any case increment the sigma index counter
         self.__sigma_index += 1
 
         return z
 
     def _krige(self, idx):
-        """Algorithm
+        r"""Algorithm
 
         Kriging algorithm for one point. This is the place, where the
         algorithm shall be changed and optimized.
@@ -363,8 +396,8 @@ class OrdinaryKriging:
 
         .. math::
             \hat{Z} = \sum_i(w_i * z_i)
-        
-        where :math:`w_i` is the calulated kriging weight for the i-th point 
+
+        where :math:`w_i` is the calulated kriging weight for the i-th point
         and :math:`z_i` is the observed value at that point.
 
         The kriging variance :math:`\sigma^2` (sigma) is calculate as follows:
@@ -372,8 +405,8 @@ class OrdinaryKriging:
         .. math::
             \sigma^2 = \sum_i(w_i * \gamma(p_0 - p_i)) + \lambda
 
-        where :math:`w_i` is again the weight, :math:`\gamma(p_0 - p_i)` is 
-        the semivairance of the distance between the unobserved location and 
+        where :math:`w_i` is again the weight, :math:`\gamma(p_0 - p_i)` is
+        the semivairance of the distance between the unobserved location and
         the i-th observation. :math:`\lamda` is the Lagrange multiplier needed
         to minimize the estimation error.
 
@@ -385,44 +418,32 @@ class OrdinaryKriging:
             kriging variance :math:`\sigma^2` for p.
 
         """
-        
+
         if self.perf:
             t0 = time.time()
 
-        p = self.transform_coordinates[idx,:]
-        dists = self.transform_dists[idx,:]
-        
-        # find all points within the search distance
-        idx = np.where(dists <= self.range)[0]
+        # get the point and index
+        p = self.transform_coords.coords[idx, :]
+        idx = self.transform_coords_pair.find_closest(
+            idx,
+            self.range,
+            self._maxp
+        )
 
         # raise an error if not enough points are found
         if idx.size < self._minp:
             raise LessPointsError
 
-        if idx.size > self._maxp:
-            sorted_idx = np.argsort(dists)
-            idx = sorted_idx[np.isin(sorted_idx, idx)][:self._maxp]
-
         # finally find the points and values
-        in_range = self.coords[idx]
+        in_range = self.coords.coords[idx]
         values = self.values[idx]
-        dist_mat = self.dist(in_range)
+        dist_mat = self.coords.diagonal(idx)
 
         # if performance is tracked, time this step
         if self.perf:
             t1 = time.time()
             self.perf_dist.append(t1 - t0)
 
-        # OLD ALGORITHM
-        # a = np.ones((len(in_range) + 1, len(in_range) + 1))
-
-        # fill; TODO: this can be done faster
-        #for i in range(len(in_range)):
-        #    for j in range(len(in_range)):
-        #        a[i, j] = self.V.compiled_model(dist_mat[i ,j])
-        # the outermost elements are all 1, except the last one
-        #a[-1, -1] = 0
-
         # build the kriging Matrix; needs N + 1 dimensionality
         if self.mode == 'exact':
             a = self._build_matrix(dist_mat)
@@ -449,7 +470,7 @@ class OrdinaryKriging:
 
         # solve the system
         try:
-            l = self._solve(a, b)
+            _lambda = self._solve(a, b)
         except LinAlgError as e:
             print(a)
             if str(e) == 'Matrix is singular.':
@@ -473,12 +494,12 @@ class OrdinaryKriging:
                 self.perf_solv.append(t3 - t2)
 
         # calculate Kriging variance
-        # sigma is the weights times the semi-variance to p0 
-        # plus the lagrange factor 
-        sigma = sum(b[:-1] * l[:-1]) + l[-1]
+        # sigma is the weights times the semi-variance to p0
+        # plus the lagrange factor
+        sigma = sum(b[:-1] * _lambda[:-1]) + _lambda[-1]
 
         # calculate Z
-        Z = l[:-1].dot(values)
+        Z = _lambda[:-1].dot(values)
 
         # return
         return Z, sigma
